/**
 *	Copyright (c) 2008,  Eckhart Arnold
 *
 * Dieses Programm ist freie Software. Sie können es unter den
 * Bedingungen der GNU General Public License, wie von der Free
 * Software Foundation veröffentlicht, weitergeben und/oder
 * modifizieren, entweder gemäß Version 3 der Lizenz oder (nach
 * Ihrer Option) jeder späteren Version.
 *
 * Die Veröffentlichung dieses Programms erfolgt in der Hoffnung,
 * daß es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE
 * GARANTIE, sogar ohne die implizite Garantie der MARKTREIFE oder
 * der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Details finden
 * Sie in der GNU General Public License.
 *
 * Sie sollten ein Exemplar der GNU General Public License zusammen
 * mit diesem Programm erhalten haben. Falls nicht, siehe
 * <http://www.gnu.org/licenses/>.
 */

package evolution.examples;

import org.eclipse.swt.graphics.*;


public class TSP implements ISimulation {
	final TSPData	data;
	StringBuffer log = new StringBuffer();
	IGenome[]	population = null;
	int			indexBest = -1;
	int[]		pathBest = null;
	double		bestPerformance, worstPerformance;
	double[]	fitness = null;
	long		generationCounter = 0;
	EAUtil		util = new EAUtil();

	public TSP(TSPData data) {
		this.data = data;		
	}

	private void evaluate() {
		double val, sum;
		double bestPerformance = Double.MAX_VALUE;
		double worstPerformance = 0.0;
		for (int i = 0; i < population.length; i++) {
			val = population[i].roundTrip();
			if (val < bestPerformance) {
				indexBest = i;
				bestPerformance = val;
			}
			if (val > worstPerformance)  worstPerformance = val;
			fitness[i] = 1.0/val;
		}		
		if (data.trimming) {
			double trim = 1.0/(worstPerformance*101.0/100.0);
			for (int i = 0; i < population.length; i++) fitness[i] -= trim;
		}		
		if (data.steepening) {
			for (int i = 0; i < population.length; i++) {
				fitness[i] *= fitness[i];  fitness[i] *= fitness[i]; }
		}
		if (data.boltzmann) {
			sum = 0.0;  
			double T = 1.0 / Math.min(Math.sqrt(generationCounter)+100, 500);
			// double T = 1.0 / Math.min(generationCounter / 100.0 + 100.0, 500.0);
			for (double f: fitness) sum += Math.exp(f/T);
			for (int i = 0; i < population.length; i++) 
				fitness[i] = Math.exp(fitness[i]/T) / sum;
		}	
		sum = 0.0; for (double f: fitness) sum += f;
		for (int i = 0; i < population.length; i++)  fitness[i] /= sum;		
		synchronized(this) {
			this.pathBest = population[indexBest].getPath();
			assert (!data.elite) || this.bestPerformance >= bestPerformance;
			this.bestPerformance = bestPerformance;
			this.worstPerformance = worstPerformance;
		}
	}
	
	private void select() {
		IGenome[] nextGen = new IGenome[population.length];
		EAUtil.Roulette roulette = util.new Roulette(fitness);
		if (data.universal) { 
			int index, saveIndex = -1;
			for (int i = 0; i < population.length; i++) {
				index = roulette.next();
				nextGen[i] = population[index].clone();
				if (index == indexBest) saveIndex = i;	
			}		
			assert saveIndex >= 0;
			indexBest = saveIndex;
		} else {
			int start;
			if (data.elite) {
				nextGen[0] = population[indexBest].clone();
				indexBest = 0;
				start = 1;
			} else {
				indexBest = -1;
				start = 0;
			}
			for (int i = start; i < population.length; i++)
				nextGen[i] = population[roulette.pick()].clone();
		}
		population = nextGen;

	}

	private void crossover() {
		EAUtil.RandomIndices selector = util.new RandomIndices(population.length);
		int a, b, from, to;
		while (selector.left() >= 2) {
			a = selector.next();
			b = selector.next();
			if ((!data.elite) || (a != indexBest && b != indexBest)) {
				if (EAUtil.rnd.nextDouble() < data.pCrossover) {
					from = EAUtil.rnd.nextInt(data.points);
					to = EAUtil.rnd.nextInt(data.points-1);
					if (to == from) to = data.points-1;
					population[a].exchange(population[b], from, to);
					population[a].repair();
					population[b].repair();
				}
			}
		}
	}

	private void mutation() {
		IGenome g;
		for (int i = 0; i < population.length; i++) {
			if ((!data.elite) || (i != indexBest)) {
				g = population[i];
				for (int a = 0; a < g.getLength(); a++) {
					if (EAUtil.rnd.nextDouble() < data.pMutation) {
						g.mutate(a);
					}
				}
			}
		}	
	}

	// ISimulation interface
	
	public StringBuffer accessLog() {
		return log;
	}

	public void drawState(GC gc) {
		int[] path;
		double bestPerformance;
		Rectangle r = gc.getClipping();
		synchronized(this) { 
			path = this.pathBest;  
			bestPerformance = this.bestPerformance; 
		}
		gc.fillRectangle(r);
		if (path == null) {
			for (int i = 0; i < data.points; i++) {
				gc.drawPoint((int) (data.location[i][0] * r.width + 0.5),
						     (int) (data.location[i][1] * r.height + 0.5));
			}
		} else {
			int pos, x1, y1, x2, y2;
			pos = path[path.length-1];
			x1 = (int) (data.location[pos][0] * r.width + 0.5);
			y1 = (int) (data.location[pos][1] * r.height + 0.5);
			for (int i = 0; i < path.length; i++) {
				pos = path[i];
				x2 = (int) (data.location[pos][0] * r.width + 0.5);
				y2 = (int) (data.location[pos][1] * r.height + 0.5);
				gc.drawLine(x1, y1, x2, y2);
				x1 = x2;  y1 = y2;
			}
		}

		String msg;
		msg = "Generation: " + generationCounter + "  Performance: " + 
		      (new Double(bestPerformance*100)).longValue();
		Point e = gc.stringExtent(msg);
		gc.drawString(msg, r.x+r.width-e.x, r.y+r.height-e.y, false);
	}
	
	private final String BEST_PERFORMANCE = "best performance";
	private final String WORST_PERFORMANCE = "worst performance";
	
	public String[] queryMonitors() {
		String[] monitors = { BEST_PERFORMANCE, WORST_PERFORMANCE };
		return monitors;
	}
	
	public double readMonitor(String monitor) {
		if (monitor == BEST_PERFORMANCE) return bestPerformance;
		else if (monitor == WORST_PERFORMANCE) return worstPerformance;
		else {
			assert false;
			return 0.0;
		}
	}

	public boolean finished() {
		return false;
	}

	public void terminate() {
		population = null;
		fitness = null;
	}
	
	public String getDescription() {
		StringBuffer flags = new StringBuffer();
		if (data.trimming)
			flags.append("- Verschiebung der Fitnesswerte (so dass der niedrigste "+
					     "Fitnesswert annährend bei Null liegt)\n");
		if (data.steepening)
			flags.append("- Vergrößerung der Distanzen zwischen den Fitnesswerten (Skalierung)\n");
		if (data.boltzmann)
			flags.append("- Boltzmann Selektion (\"simulierte Abkühlung\")\n");
		if (data.elite)
			flags.append("- Elite Prinzip: Ein Exemplar des besten Genoms überlebt immer.\n");
		if (data.universal)
			flags.append("- Gleichmäßige (\"universelle\") Selektion "+
					     "(weniger zufällig als \"Rouletterad-Selektion\").\n");
		else flags.append("- Rouletterad Selektion\n");
		if (data.pointgenome) 
			flags.append("- Genom repräsentiert die Folge der Punkte.");
		else flags.append("- Genom repräsentiert die Folge der jeweils verbleibenden Orte.");
		return "Das \"Problem des Handlungsreisenden\" ist ein klassisches "+
		       "Problem der Informatik. Es geht darum, eine möglichst kurze "+
		       "Strecke zu finden, die eine gegebene Menge von Punkten "+
		       "(\"Orten\", die besucht werden müssen) verbindet und "+
		       "- in der hier vorgestellten Variante - wieder zum "+
		       "Ausgangspunkt zurück führt.\n\n "+
		       "Hier soll anhand dieses Problems vor Augen geführt werden "+
		       "wie evolutionäre Algorithmen funktionieren. Das gewählte "+
		       "Beispiel arbeitet mit folgenden Parametern: \n\n"+
		       "Anzahl der Orte: "+data.points+"\n" + 
		       "Häufungspunkte: "+data.clusters+"\n"+
		       "Populationsgröße: "+data.popSize+"\n"+
		       "Rekombination: "+(new Double(data.pCrossover*100)).intValue()+" %\n"+
		       "Mutationen: "+
		       (new Double(data.pMutation*data.points*100)).intValue()+" %\n"+
		       "\nWeitere Einstellungen:\n"+
		       flags.toString();
	}

	public void reset() {
		log = new StringBuffer();
		population = new IGenome[data.popSize];
		fitness = new double[data.popSize];
		indexBest = -1;
		pathBest = null;
		bestPerformance = Double.MAX_VALUE;
		worstPerformance = 0.0;
		generationCounter = 0;
		for (int i = 0; i < population.length; i++)
			if (data.pointgenome) population[i] = new PointGenome(data);
			else population[i] = new TailGenome(data);
	}

	public void step() {
		if (population == null) 
			throw new Error("Simulation not initialized. Call reset() first!"); 
		evaluate();
		select();
		crossover();
		mutation();
		generationCounter += 1;

		double e = Math.log10((double) (generationCounter));
		int i = (int) Math.pow(10, Math.floor(e));
		if (generationCounter % i == 0) {
			log.append("Generation: " + generationCounter +
					   "   Best: " +
					   (new Double(bestPerformance*100)).intValue() +
					   "   Worst: " +
					   (new Double(worstPerformance*100)).intValue() +
					   "\n");		
		}
	}
}
